<!doctype html>
<html class="dark">

<head>
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <script src="https://cdn.tailwindcss.com"></script>
  <script>
    // Toggle dark mode based on user's system preference
    if (window.matchMedia && window.matchMedia('(prefers-color-scheme: dark)').matches) {
      document.documentElement.classList.add('dark');
    } else {
      document.documentElement.classList.remove('dark');
    }
  </script>
</head>

<body class="bg-white dark:bg-gray-900 text-gray-900 dark:text-gray-100">
  <header class="bg-white dark:bg-gray-800 text-gray-900 dark:text-gray-100 p-4 shadow-md">
    <div class="container mx-auto flex justify-between items-center">
      <img src="/static/logo.png" alt="AutoTrain" class="w-32">
    </div>
  </header>

  <div class="form-container max-w-lg mx-auto mt-10 p-6 shadow-2xl bg-white dark:bg-gray-800 rounded-lg">
    <h1 class="text-3xl font-bold mb-6">Error</h1>
    <p class="text-red-500 text-lg mb-6">HF_TOKEN environment variable is not set.</p>
    <a href="/" class="text-blue-500 dark:text-blue-400 hover:underline">Go back to Home</a>
  </div>
</body>

</html>